import errno
import os.path
import base64
import sys
from pathlib import Path
from argparse import _ArgumentGroup, Namespace
from frida_fusion.libs.logger import Logger
from frida_fusion.libs.scriptlocation import ScriptLocation
from frida_fusion.module import ModuleBase


class HermesJSInjector(ModuleBase):

    def __init__(self):
        super().__init__('Hermes Injector', 'Inject JavaScript at Hermes engine used by React Native.')
        self.mod_path = str(Path(__file__).resolve().parent)
        self.js_file = os.path.join(self.mod_path, "hermes_injector.js")
        self.hermes_js_file = ""
        self._suppress_messages = False

    def start_module(self, **kwargs) -> bool:
        pass

    def js_files(self) -> list:
        return [
            self.js_file
        ]

    def suppress_messages(self):
        self._suppress_messages = True

    def dynamic_script(self) -> str:
        return f"const FUSION_HERMES_JS = '{self.hermes_js_file}';"
    
    def add_params(self, flags: _ArgumentGroup):
        flags.add_argument('--hermes-js-script',
                           dest='hermes_js_file',
                           metavar='file',
                           default=None,
                           required=True,
                           type=str,
                           help='Hermes JS file path')

    def load_from_arguments(self, args: Namespace) -> bool:
        if args.hermes_js_file is None:
            return False

        try:
            b_data = bytearray()
            with open(os.path.join(self.mod_path, "hermes_hook.js"), 'rb') as f:
                b_data = bytearray(f.read())

            if len(b_data) > 0:
                b_data += bytearray([0x0a, 0x0a])

            with open(args.hermes_js_file, 'rb') as f:
                b_data += bytearray(f.read())
            
            self.hermes_js_file = base64.b64encode(b_data).decode("UTF-8")

        except IOError as x:
            if x.errno == errno.EACCES:
                Logger.pl('{!} {R}error: could not open hermes js file {O}permission denied{R}{W}\r\n')
                sys.exit(1)
            elif x.errno == errno.EISDIR:
                Logger.pl('{!} {R}error: could not open hermes js file {O}it is an directory{R}{W}\r\n')
                sys.exit(1)
            else:
                Logger.pl('{!} {R}error: could not open hermes js file: %s{W}\r\n' % str(x))
                sys.exit(1)

        return True

    def key_value_event(self,
                        script_location: ScriptLocation = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:
        return True

    def data_event(self,
                   script_location: ScriptLocation = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        return True






class SilentKillError(Exception):
    def __init__(self, *args, **kwargs):
        # args -> vão para a Exception (ex.: a mensagem)
        # kwargs -> metadados opcionais seus
        self.meta = kwargs
        self._msg = kwargs.get("message") or ""
        if args is not None and len(args) > 0 and self._msg == "":
            self._msg = args[0]
        super().__init__(*args)

    def __str__(self):
        if self._msg != "":
            return self._msg
        else:
            return str(super(SilentKillError, self).__str__())

    def __repr__(self):
        return str(self)



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

import argparse
import sys

from .libs.color import Color
from .__meta__ import __description__
from .module import ModuleManager


class Arguments(object):
    ''' Holds arguments used by the Frida Fusion '''
    restore = False

    def __init__(self, custom_args=''):
        self.verbose = any(['-v' in word for word in sys.argv])
        self.restore = any(['-R' in word for word in sys.argv])
        self.args = self.get_arguments(custom_args)

    def _verbose(self, msg):
        if self.verbose:
            return Color.s(msg)
        else:
            return argparse.SUPPRESS

    def get_arguments(self, custom_args=''):
        ''' Returns parser.args() containing all program arguments '''

        parser = argparse.ArgumentParser(
            usage=argparse.SUPPRESS,
            formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=80,
                                                                width=130)
        )

        device_group = parser.add_argument_group('Device selector')
        self._add_device_args(device_group)

        app_group = parser.add_argument_group('Application selector')
        self._add_app_args(app_group)

        glob = parser.add_argument_group('General Setting')
        self._add_global_args(glob)

        modules_group = parser.add_argument_group('Modules')
        self._add_modules_args(modules_group)

        # Add module args
        for mod in self._get_requested_module_list():
            flags = parser.add_argument_group(f'{mod.name} Flags')
            mod.create_instance().add_params(flags)

        return parser.parse_args()

    def _get_requested_module_list(self):
        mods = ModuleManager.list_modules()
        parser = argparse.ArgumentParser()
        parser.add_argument('-m',
                            '--module',
                            action='append',
                            dest='enabled_modules')

        t_args, _ = parser.parse_known_args([
            word
            for word in sys.argv
            if word != "-h" and word != "--list-modules"
        ])
        if t_args is None or t_args.enabled_modules is None:
            return []
        return [
            m
            for md in t_args.enabled_modules
            for mn in md.split(",")
            if mn.strip() != ""
            for _, m in mods.items()
            if m.safe_name() == mn.lower()
        ]

    def _add_app_args(self, app):
        app.add_argument('-f',
                         '--package',
                         dest='app_id',
                         metavar='[APP ID]',
                         required=False,
                         default=None,
                         help='Spawn application ID')

        app.add_argument('-p',
                         '--attach-pid',
                         dest='pid',
                         metavar='[PID]',
                         default=0,
                         required=False,
                         type=int,
                         help='Spawn application ID')

    def _add_device_args(self, device):
        device.add_argument('-D',
                            '--device',
                            dest='device_id',
                            metavar='[ID]',
                            type=str,
                            required=False,
                            default=None,
                            help='Connect to device with the given ID')

        device.add_argument('-U',
                            '--usb',
                            action='store_true',
                            dest='use_usb',
                            default=False,
                            required=False,
                            help='Connect to USB device')

        device.add_argument('-R',
                            '--remote',
                            action='store_true',
                            dest='use_remote',
                            default=False,
                            required=False,
                            help='Connect to remote frida-server')

        device.add_argument('-H',
                            '--host',
                            dest='remote_host',
                            metavar='[HOST]',
                            type=str,
                            required=False,
                            default=None,
                            help='Connect to remote frida-server on HOS')

    def _add_global_args(self, glob):
        glob.add_argument('-s',
                          '--script-path',
                          dest='frida_scripts',
                          default=None,
                          metavar='[path]',
                          type=str,
                          help='JS File path or directory with Frida script')

        glob.add_argument('--delay-injection',
                          action='store_true',
                          dest='delay',
                          default=False,
                          help='Delay script injection')

        glob.add_argument('--show-time',
                          action='store_true',
                          dest='show_time',
                          default=False,
                          help='Display time')

        glob.add_argument('-o',
                          action='store',
                          dest='out_file',
                          metavar='[output file]',
                          type=str,
                          help=Color.s('Save output to disk (default: {G}none{W})'))

        glob.add_argument('-l',
                          '--min-level',
                          action='store',
                          dest='debug_level',
                          metavar='[level]',
                          type=str,
                          default='I',
                          help=Color.s('Minimum log level to be displayed (V,D,I,W,E,F) (default: {G}I{W})'))

    def _add_modules_args(self, modules):
        modules.add_argument('--list-modules',
                             action='store_true',
                             default=False,
                             dest='list_modules',
                             help=Color.s('List available modules'))

        modules.add_argument('-m',
                             '--module',
                             action='append',
                             dest='enabled_modules',
                             help='Enabled module by name. You can specify multiple values repeating the flag.')

        modules.add_argument('--silence-module-messages',
                             action='append',
                             dest='ignore_messages_modules',
                             help='Silence messages becaming from module by name. You can specify multiple values repeating the flag.')




#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import base64
import inspect
import json
import datetime
import re
from pathlib import Path
from typing import Optional

from .scriptlocation import ScriptLocation
from ..libs.color import Color


class Logger(object):
    ''' Helper object for easily printing colored text to the terminal. '''

    out_file = ''
    debug_level = 0
    print_timestamp = True
    filename_col_len = 0

    level_map = {
        "V": 0,
        "D": 1,
        "I": 2,
        "W": 3,
        "E": 4,
        "F": 5,
    }

    level_tag = {
        "*": "D",
        "+": "I",
        "-": "W",
    }

    @staticmethod
    def pl(text):
        '''Prints text using colored format with trailing new line.'''
        Color.pl(text)

        if Logger.out_file != '':
            try:
                with open(Logger.out_file, "a") as text_file:
                    text_file.write(Color.escape_ansi(Color.sc(text)) + '\n')
            except:
                pass

    @staticmethod
    def pl_file(text):
        '''Prints text using colored format with trailing new line.'''

        if Logger.out_file != '':
            try:
                with open(Logger.out_file, "a") as text_file:
                    text_file.write(Color.escape_ansi(Color.sc(text)) + '\n')
            except:
                Color.pl(text)
        else:
            Color.pl(text)

    @classmethod
    def get_caller_info(cls, stack_index: int = 1) -> ScriptLocation:
        """Retrieves information about the calling script, function, and line number."""
        # inspect.stack() returns a list of frame records.
        # Each frame record is a tuple containing:
        # (frame object, filename, line number, function name, list of lines of context, index of current line in context)

        # The first element (index 0) is the current function (get_caller_info).
        # The second element (index 1) is the caller of get_caller_info.
        caller_frame = inspect.stack()[stack_index]

        # Extract information from the caller's frame
        filename = caller_frame.filename
        line_number = caller_frame.lineno
        function_name = caller_frame.function

        # Optionally, get the base name of the script for cleaner output
        script_name = Path(filename).name

        return ScriptLocation(
            file_name=script_name,
            function_name=function_name,
            line=str(line_number)
        )

    @classmethod
    def get_error_info_from_format_exc(cls, stack_index: int = -1) -> Optional[ScriptLocation]:
        """
        Faz o *parse* do texto gerado por traceback.format_exc() e extrai arquivo/linha/função.
        - frame_index: -1 pega o último frame (onde a exceção estourou).
        Formato esperado das linhas:
            File "/caminho/mod.py", line 123, in func
        """
        from traceback import format_exc

        # Captura tuplas (arquivo, linha, função)
        # Obs.: tolera espaços e caminhos com aspas; não captura a linha de código em si.
        pattern = r'(?i:File) "(.+?)", (?i:line) (\d+), (?i:in) ([^\n\r]+)'
        matches = re.findall(pattern, format_exc())

        if not matches:
            return None

        file_path, lineno, func = matches[stack_index]
        return ScriptLocation(
            file_name=Path(file_path).name,
            function_name=func.strip(),
            line=str(lineno),
        )

    @staticmethod
    def json_serial(obj):
        """JSON serializer for objects not serializable by default json code"""

        if isinstance(obj, (datetime.datetime, datetime.date)):
            # obj = obj.astimezone(datetime.timezone(datetime.timedelta(hours=0), 'Z'))
            return obj.strftime("%Y-%m-%dT%H:%M:%S.000Z")

        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode("UTF-8")

        try:
            return str(obj)
        except:
            return f"[ERROR] Type %s not serializable{type(obj)}"

    @classmethod
    def check_print(cls, level: str = "*"):

        if level is None:
            level = "*"

        dbg_tag = next(iter([
            k
            for k in Logger.level_map.keys()
            if level.upper() == k
        ]), Logger.level_tag.get(level, "I"))

        dbg_idx = Logger.level_map.get(dbg_tag, 0)

        return Logger.debug_level <= dbg_idx

    @classmethod
    def print_message(cls, level: str = "*", message: str = "",
                      script_location: ScriptLocation = None, filename_col_len: int = 26):

        if level is None:
            level = "*"

        dbg_tag = next(iter([
            k
            for k in Logger.level_map.keys()
            if level.upper() == k
        ]), Logger.level_tag.get(level, "I"))

        dbg_idx = Logger.level_map.get(dbg_tag, 0)

        if Logger.debug_level > dbg_idx:
            return

        if filename_col_len < 26:
            filename_col_len = 26

        if Logger.filename_col_len > filename_col_len:
            filename_col_len = Logger.filename_col_len

        prefix = ""
        if Logger.print_timestamp:
            ts = datetime.datetime.now()
            stamp = f"{ts:%H:%M:%S}.{int(ts.microsecond / 1000):03d}"
            prefix += f"\033[2m{stamp.ljust(13)}{Color.color_reset}"

        fg_color = Color.color_level[dbg_idx]
        tag_color = Color.color_tags[dbg_idx]

        if script_location is None:
            script_location = cls.get_caller_info(stack_index=2)

        if script_location.file_name == "frida/node_modules/frida-java-bridge/lib/class-factory.js":
            file_name = "frida/.../class-factory.js"
        else:
            file_name = str(Path(script_location.file_name).name)

        if len(file_name) > filename_col_len:
            file_name = f"{file_name[0:filename_col_len-3]}..."

        prefix += (f"{fg_color}{file_name.rjust(filename_col_len)}"
                   f"{Color.color_reset}\033[2m:{str(script_location.line).ljust(10)}"
                   f"{Color.color_reset} ")
        prefix_len = len(Color.escape_ansi(prefix))

        f_message = ""
        if message is None:
            message = ""

        if isinstance(message, dict):
            try:
                message = json.dumps(message, ident=4, default=Logger.json_serial)
            except:
                message = str(message)

        for line in message.split("\n"):
            if f_message == "":
                f_message += (f"{prefix}{tag_color} {dbg_tag} {Color.color_reset} "
                              f"{fg_color}{line}{Color.color_reset}")
            else:
                f_message += (f"\n{''.rjust(prefix_len)}{tag_color} {dbg_tag} "
                              f"{Color.color_reset} {fg_color}{line}{Color.color_reset}")

        Logger.pl(f_message)

    @classmethod
    def print_exception(cls, err):
        from traceback import format_exc
        err_txt = 'Error:{O} %s{W}' % str(err)
        err_txt += '\n{O}Full stack trace below\n'
        err_txt += format_exc().strip()

        err_txt = err_txt.replace('\n', '\n{W}   ')
        err_txt = err_txt.replace('  File', '{W}{D}File')
        err_txt = err_txt.replace('  Exception: ', '{R}Exception: {O}')

        Logger.pl(f"{err_txt}\n")



import base64
import errno
import os.path
import json
from pathlib import Path
from frida_fusion.libs.logger import Logger
from frida_fusion.module import ModuleBase
from frida_fusion.libs.scriptlocation import ScriptLocation


class OkHttpLogging(ModuleBase):

    def __init__(self):
        super().__init__('OkHttp3 logging', 'Use okhttp-logging')
        self.mod_path = str(Path(__file__).resolve().parent)
        self.js_file = os.path.join(self.mod_path, "okhttp-logging.js")
        self._suppress_messages = False
        self._log_level = 'BODY'

    '''
    const Level = {
          NONE: 'NONE',
          BASIC: 'BASIC',
          HEADERS: 'HEADERS',
          BODY: 'BODY',
          STREAMING: 'STREAMING'
        };
    '''

    def start_module(self, **kwargs) -> bool:
        pass

    def js_files(self) -> list:
        return [
            self.js_file
        ]

    def suppress_messages(self):
        self._suppress_messages = True

    def dynamic_script(self) -> str:
        return f"const FF_OKHTTP_LOGGING_LEVEL = '{self._log_level}';"

    def key_value_event(self,
                        script_location: ScriptLocation = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:

        if module == "okhttp!intercept":
            if not self._suppress_messages:
                Logger.print_message(
                    level="I",
                    message=self.format_output(received_data),
                    script_location=script_location
                )

        elif module == "okhttp!intercept!interceptors":
            if not self._suppress_messages:
                t_name = received_data.get('type', 'interceptor')
                i_class = received_data.get('interceptorclass', '')
                Logger.print_message(
                    level="I",
                    message=f"okhttp {t_name} found!\nClass: {i_class}",
                    script_location=script_location
                )
            
        return True

    def data_event(self,
                   script_location: ScriptLocation = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        return True

    @classmethod
    def format_output(cls, received_data: dict = None) -> str:
        unk = "<unknown>"
        error = received_data.get("error", None)
        response_code = received_data.get("response-code", None)
        elapsed_ime = received_data.get("elapsed-time", "")
        method = received_data.get("method", unk)
        url = received_data.get("url", unk)
        protocol = (received_data.get("protocol", unk) or unk).upper()
        headers = received_data.get("request-header", "").replace("\r", "")
        data = f'--> {method} {url} {protocol}\n'
        data += f'{headers}\n'
        request_body = received_data.get("request-body", None)
        try:
            if request_body is not None and request_body.strip() != "":
                request_body = base64.b64decode(request_body.encode("UTF-8")).decode("UTF-8")
            else:
                request_body = None
        except Exception:
            pass

        if request_body is not None:
            data += f'{request_body}\n\n'

        if error is not None:
            data += f'<-- HTTP ERROR ({elapsed_ime}ms)\n'
            data += f'{error}\n'

        if response_code is not None:
            response_status = received_data.get("response-status", None)
            response_body = received_data.get("response-body", None)
            response_header = received_data.get("response-header", "")
            try:
                if response_body is not None and response_body.strip() != "":
                    response_body = base64.b64decode(response_body.encode("UTF-8")).decode("UTF-8")
                else:
                    response_body = None
            except Exception:
                pass

            # Try to decode as json
            try:
                response_body = json.dumps(
                    json.loads(response_body),
                    default=Logger.json_serial,
                    indent=4, sort_keys=False)
            except Exception:
                pass

            data += f'<-- {response_code} {response_status} ({elapsed_ime}ms)\n'
            data += f'{response_header}\n'
            if response_body is not None:
                data += f'{response_body}\n\n'

        data += f'<-- END HTTP\n'

        return data



import json
import os.path
from pathlib import Path
from frida_fusion.libs.logger import Logger
from frida_fusion.libs.database import Database
from frida_fusion.libs.scriptlocation import ScriptLocation
from frida_fusion.module import ModuleBase


class SharedPreferences(ModuleBase):
    _hide_commons = [
        "RCTI18nUtil",
        "NETWORK_USAGE_TRACKING_",
    ]

    class SharedPreferencesDB(Database):
        dbName = ""

        def __init__(self, db_name: str):
            super().__init__(
                auto_create=True,
                db_name=db_name
            )
            self.create_db()

        def create_db(self):
            super().create_db()
            conn = self.connect_to_db(check=False)

            # definindo um cursor
            cursor = conn.cursor()

            # criando a tabela (schema)
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS [shared_preferences] (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    package TEXT NOT NULL,
                    module TEXT NULL,
                    class_name TEXT NULL,
                    method TEXT NULL,
                    args TEXT NULL,
                    result TEXT NULL,
                    stack_trace TEXT NULL,
                    created_date datetime not null DEFAULT (datetime('now','localtime'))
                );
            """)

            conn.commit()

            # Must get the constraints
            self.get_constraints(conn)

    def __init__(self):
        super().__init__('Shared Preferences', 'Hook android.content.SharedPreferences functions')
        self._shared_preferences_db = None
        self._package = None
        self._suppress_messages = False
        self.mod_path = str(Path(__file__).resolve().parent)

    def start_module(self, **kwargs) -> bool:
        if 'db_path' not in kwargs:
            raise Exception("parameter db_path not found")

        self._package = kwargs['package']
        self._shared_preferences_db = SharedPreferences.SharedPreferencesDB(db_name=kwargs['db_path'])
        return True

    def js_files(self) -> list:
        return [
            os.path.join(self.mod_path, "shared_preferences.js")
        ]

    def suppress_messages(self):
        self._suppress_messages = True

    def key_value_event(self,
                        script_location: ScriptLocation = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:

        if 'SharedPreferences.' in module:
            try:
                received_data['args'] = json.loads(received_data.get('args', "[]"))
            except Exception:
                pass
            try:
                received_data['result'] = json.loads(received_data.get('result', ""))
            except Exception:
                pass

            s_args = received_data.get('args', "[]")
            if not isinstance(s_args, str):
                s_args = json.dumps(s_args, default=Logger.json_serial)

            s_result = received_data.get('result', "")
            if not isinstance(s_result, str):
                s_result = json.dumps(s_result, default=Logger.json_serial)

            class_name = received_data.get('class', '<unknown>')
            s_method = received_data.get('method', '<unknown>')

            self._shared_preferences_db.insert_one(
                table_name='shared_preferences',
                package=self._package,
                module=module,
                class_name=class_name,
                method=s_method,
                args=s_args,
                result=s_result,
                stack_trace=stack_trace
            )

        if module in [
            "SharedPreferences.getString",
            "SharedPreferences.getInt",
            "SharedPreferences.getLong",
            "SharedPreferences.getFloat",
            "SharedPreferences.getStringSet",
            "SharedPreferences.putString",
            "SharedPreferences.putInt",
            "SharedPreferences.putLong",
            "SharedPreferences.putFloat",
            "SharedPreferences.putStringSet",
        ]:
            if self._check_show(received_data):
                class_name = received_data.get('class', '<unknown>')
                s_method = received_data.get('method', '<unknown>')

                l_args = ', '.join([
                    "null" if v is None else (str(v) if isinstance(v, float) else (
                        str(v) if isinstance(v, int) else f"'{str(v)}'" if isinstance(v, str) else f"{str(v)}"))
                    for v in received_data.get('args', [])
                ])
                data = "null"
                try:
                    data = json.dumps(received_data.get('result', None), default=Logger.json_serial, indent=4,
                                      sort_keys=False)
                except Exception:
                    pass
                Logger.print_message(
                    level="I",
                    message=f"{class_name}->{s_method}({l_args})\nResult: {data}",
                    script_location=script_location
                )

        if 'SharedPreferences.' in module:
            if self._check_show(received_data):
                data = json.dumps(received_data, default=Logger.json_serial, indent=4, sort_keys=False)
                Logger.print_message(
                    level="D",
                    message=f"{module}:\n{data}\n{stack_trace}",
                    script_location=script_location
                )

        return True

    def data_event(self,
                   script_location: ScriptLocation = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        # Nothing by now
        return True

    def _check_show(self, received_data: dict = None) -> bool:
        if self._suppress_messages:
            return False

        c_args = received_data.get('args', "[]")
        if not isinstance(c_args, str):
            c_args = json.dumps(c_args, default=Logger.json_serial)

        return not any([
            True
            for k in SharedPreferences._hide_commons
            if k in c_args
        ])



import json
import os.path
from pathlib import Path
import base64
import string

from frida_fusion.libs.logger import Logger
from frida_fusion.libs.database import Database
from frida_fusion.libs.scriptlocation import ScriptLocation
from frida_fusion.module import ModuleBase


class Crypto(ModuleBase):
    class CryptoDB(Database):
        dbName = ""

        def __init__(self, db_name: str):
            super().__init__(
                auto_create=True,
                db_name=db_name
            )
            self.create_db()

        def create_db(self):
            super().create_db()
            conn = self.connect_to_db(check=False)

            # definindo um cursor
            cursor = conn.cursor()

            # criando a tabela (schema)
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS [crypto] (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    package TEXT NOT NULL,
                    hashcode TEXT NOT NULL,
                    algorithm TEXT NULL,
                    init_key TEXT NULL,
                    iv TEXT NULL,
                    flow TEXT NULL,
                    key TEXT NULL,
                    clear_text TEXT NULL,
                    clear_text_b64 TEXT NULL,
                    cipher_data TEXT NULL,
                    status TEXT NULL DEFAULT ('open'),
                    stack_trace TEXT NULL,
                    created_date datetime not null DEFAULT (datetime('now','localtime'))
                );
            """)

            conn.commit()

            cursor.execute("""
                CREATE TABLE IF NOT EXISTS [crypto_key] (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    package TEXT NOT NULL,
                    key TEXT NULL,
                    printable_key TEXT NULL,
                    salt TEXT NULL,
                    iteration_count INTEGER NULL DEFAULT (0),
                    key_class TEXT NULL DEFAULT ('<unknown>'),
                    additional_data TEXT NULL,
                    created_date datetime not null DEFAULT (datetime('now','localtime')),
                    UNIQUE (package, key, key_class)
                );
            """)

            conn.commit()

            cursor.execute("""
                CREATE TABLE IF NOT EXISTS [digest] (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    package TEXT NOT NULL,
                    algorithm TEXT NOT NULL,
                    hashcode TEXT NULL,
                    clear_text TEXT NULL,
                    clear_text_b64 TEXT NULL,
                    hash_b64 TEXT NULL,
                    hash_hex TEXT NULL,
                    stack_trace TEXT NULL,
                    created_date datetime not null DEFAULT (datetime('now','localtime'))
                );
            """)

            conn.commit()

            # Must get the constraints
            self.get_constraints(conn)

        @classmethod
        def get_printable(cls, b64_data):
            try:
                text = base64.b64decode(b64_data).decode("UTF-8")
                if all(c in string.printable for c in text):
                    return text
                else:
                    return ''
            except Exception as e:
                # Color.pl('{!} {R}Erro getPrintable:{O} %s{W}' % str(e))
                return ''

        def update_crypto(self, iv=None, hashcode=None, flow=None, key=None, before_final=None,
                          after_final=None, stack_trace=None, id=None, algorithm=None,
                          status=None, package=None):

            conn = self.connect_to_db(check=False)
            cursor = conn.cursor()

            select = "select id, flow from [crypto] where status = 'open'"
            data = []
            if hashcode is not None and (before_final is not None or after_final is not None):
                select += " and hashcode = ?"
                data = (hashcode,)

            if id is not None and (before_final is not None or after_final is not None):
                select += " and id = ?"
                data = (id,)

            cursor.execute(select, data)

            f = cursor.fetchall()
            if f:
                last = f[0]
                id = last[0]
                dbflow = last[1]
                integrity = False

                data = []
                update = "update [crypto] set "

                if package is not None:
                    integrity = True
                    update += " package = ?,"
                    data.append(package)

                if iv is not None:
                    integrity = True
                    update += " iv = ?,"
                    data.append(iv)

                if hashcode is not None:
                    integrity = True
                    update += " hashcode = ?,"
                    data.append(hashcode)

                if flow is not None:
                    integrity = True
                    update += " algorithm = ?,"
                    data.append(algorithm)

                if flow is not None:
                    integrity = True
                    update += " flow = ?,"
                    data.append(flow)

                if key is not None:
                    integrity = True
                    update += " key = ?,"
                    data.append(key)

                if key is not None:
                    integrity = True
                    update += " key = ?,"
                    data.append(key)

                if stack_trace is not None:
                    integrity = True
                    update += " stack_trace = ?,"
                    data.append(stack_trace)

                if status is not None:
                    integrity = True
                    update += " status = ?,"
                    data.append(status)

                if before_final is not None:
                    integrity = True
                    if dbflow == "enc":
                        update += " clear_text = ?, clear_text_b64 = ?,"
                        data.append(self.get_printable(before_final))
                    else:
                        update += " cipher_data = ?,"

                    data.append(before_final)

                if after_final is not None:
                    integrity = True
                    if dbflow == "enc":
                        update += " cipher_data = ?, status = 'complete'"

                    else:
                        update += " clear_text = ?, clear_text_b64 = ?, status = 'complete'"
                        data.append(self.get_printable(after_final))

                    data.append(after_final)

                if integrity:
                    # update += " status = 'incositente'"

                    # Em cenários onde o campo de entrada (encriptado é nulo, não vai ter nada a processar)
                    update = update.strip(",").strip()
                    update += " where id = ?"

                    data.append(id)
                    # data.append(None)

                    cursor.execute(update, data)

                    cursor = conn.cursor()
                    cursor.execute("""
                    delete from [crypto]
                    where algorithm is null and init_key is null and key is null and clear_text is null
                    and hashcode in (
                        select hashcode from [crypto]
                        where id = ?
                    )
                    """, (id,))

                    conn.commit()

                    # Color.pl('{+} {W}Crypto atualizada. {C}ID: {O}%s{W}' % id)

            conn.close()

        def insert_digest(self, package, hashcode, algorithm, data_input, data_output, stack_trace):

            conn = self.connect_to_db(check=False)

            clear_text = ""
            clear_text_b64 = ""
            if data_input is not None:

                if isinstance(clear_text, bytes):
                    clear_text = clear_text.decode("UTF-8")
                    clear_text_b64 = base64.b64encode(clear_text).decode("UTF-8")
                else:
                    clear_text_b64 = data_input
                    try:
                        clear_text = base64.b64decode(data_input).decode("UTF-8")
                    except:
                        pass

            hash_b64 = ""
            hash_hex = ""
            if data_output is not None:
                if isinstance(data_output, bytes):
                    hash_hex = ''.join('{:02X}'.format(b) for b in data_output)
                    hash_b64 = base64.b64encode(data_output).decode("UTF-8")
                else:
                    hash_b64 = data_output
                    hash_hex = ''.join('{:02X}'.format(b) for b in base64.b64decode(data_output))

            cursor = conn.cursor()
            cursor.execute("""
            insert into [digest] ([package], [hashcode], [algorithm], [clear_text], [clear_text_b64], [hash_b64], [hash_hex], [stack_trace])
            VALUES (?,?,?,?,?,?,?,?);
            """, (package, hashcode, algorithm, clear_text, clear_text_b64, hash_b64, hash_hex, stack_trace,))

            conn.commit()

            conn.close()

            # Color.pl('{+} {W}Inserindo crypto. {C}Algorithm: {O}%s{W}' % algorithm)

        def insert_crypto(self, package, hashcode, algorithm, init_key):

            if hashcode is None:
                return

            rows = self.select(
                table_name='crypto',
                package=package,
                hashcode=hashcode,
                status='open'
            )
            if len(rows) == 0 or not any(iter([
                True
                for r in rows
                if (
                    (algorithm is not None and algorithm == r['algorithm'])
                    or (r['algorithm'] is None or r['algorithm'].strip() == "")
                ) and (
                    (
                        (init_key is not None and init_key != '' and init_key != 'IA==')
                        and (init_key == r['init_key'] or init_key == r['key'])
                    )
                    or (r['init_key'] is None or r['init_key'].strip() == "")
                )
            ])):
                if init_key is not None and init_key != '' and init_key != 'IA==':
                    self.insert_one(
                        table_name='crypto',
                        package=package,
                        hashcode=hashcode,
                        algorithm=algorithm,
                        init_key=init_key,
                        status='open')
                else:
                    self.insert_one(
                        table_name='crypto',
                        package=package,
                        hashcode=hashcode,
                        algorithm=algorithm,
                        status='open')

        def insert_crypto_key(self, package, key, key_class, salt=None,
                              iteration_count=0, module="<unknown>", additional_data=dict):
            if key is not None and key != '' and key != 'IA==':
                self.insert_ignore_one(
                    table_name='crypto_key',
                    package=package,
                    key=key,
                    printable_key=self.get_printable(key),
                    key_class=key_class,
                    salt=salt,
                    iteration_count=iteration_count,
                    additional_data=json.dumps({
                        **{"module": module},
                        **(additional_data if additional_data is not None and isinstance(additional_data, dict) else {})
                    }, default=Logger.json_serial)
                )

    def __init__(self):
        super().__init__('Crypto', 'Hook cryptography/hashing functions')
        self._package = None
        self._crypto_db = None
        self._suppress_messages = False
        self.mod_path = str(Path(__file__).resolve().parent)

    def start_module(self, **kwargs) -> bool:
        if 'db_path' not in kwargs:
            raise Exception("parameter db_path not found")

        self._package = kwargs['package']
        self._crypto_db = Crypto.CryptoDB(db_name=kwargs['db_path'])
        return True

    def js_files(self) -> list:
        return [
            os.path.join(self.mod_path, "crypto.js")
        ]

    def suppress_messages(self):
        self._suppress_messages = True

    def key_value_event(self,
                        script_location: ScriptLocation = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:

        if module in ["X509EncodedKeySpec.init", "GCMParameterSpec.init", "PBEParameterSpec.init"]:

            key_class = received_data.get('classtype', module)
            salt = None
            iteration_count = 0

            key = received_data.get('key', None)
            if module == "GCMParameterSpec.init":
                key = received_data.get('iv_key', None)

            if module == "PBEParameterSpec.init":
                key = "None"
                salt = received_data.get('pbe_salt', None)
                iteration_count = received_data.get('iteration_count', None)

            self._crypto_db.insert_crypto_key(
                package=self._package,
                key=key,
                key_class=key_class,
                salt=salt,
                iteration_count=iteration_count,
                module=module,
                additional_data=received_data
            )

        if module == "SecretKeySpec.init":
            algorithm = received_data.get('algorithm', None)
            key = received_data.get('key', None)
            hashcode = received_data.get('hashcode', None)
            key_class = received_data.get('classtype', "SecretKeySpec")
            self._crypto_db.insert_crypto(
                package=self._package,
                hashcode=hashcode,
                algorithm=algorithm,
                init_key=key)

            self._crypto_db.insert_crypto_key(
                package=self._package,
                key=key,
                key_class=key_class,
                module=module,
                additional_data=received_data
            )

        elif module == "IvParameterSpec.init":
            bData = received_data.get('iv_key', None)
            key_class = received_data.get('classtype', "IvParameterSpec")

            offset = received_data.get('offset', None)
            length = received_data.get('length', None)
            if offset is not None and length is not None:
                try:
                    offset = int(offset)
                    length = int(length)
                    if isinstance(bData, str):
                        bData = base64.b64decode(bData)
                        if offset + length <= len(bData):
                            bData = base64.b64encode(bData[offset:offset+length]).decode("UTF-8")
                except:
                    pass

            self._crypto_db.update_crypto(iv=bData)

            self._crypto_db.insert_crypto_key(
                package=self._package,
                key=bData,
                key_class=key_class,
                module=module,
                additional_data=received_data
            )

        elif module == "cipher.init":
            hashcode = received_data.get('hashcode', None)
            opmode = received_data.get('opmode', "")
            key_class = received_data.get('keytype', "")
            key = received_data.get('key', None)
            algorithm = received_data.get('algorithm', None)

            self._crypto_db.insert_crypto(
                package=self._package,
                hashcode=hashcode,
                algorithm=algorithm,
                init_key=key
            )

            self._crypto_db.update_crypto(
                package=self._package,
                hashcode=hashcode,
                flow='enc' if 'encrypt' in opmode else ('dec' if 'decrypt' in opmode else str(opmode)),
                key=key,
                algorithm=algorithm
            )

            self._crypto_db.insert_crypto_key(
                package=self._package,
                key=key,
                key_class=key_class,
                module=module,
                additional_data=received_data
            )

            if not self._suppress_messages:
                Logger.print_message(
                    level="I",
                    message=f"Cipher init received\nHashcode: {hashcode}\nOpmode: {opmode}\nKeytype: {key_class}",
                    script_location=script_location
                )
        
        elif module == "cipher.getInstance":
            hashcode = received_data.get('hashcode', None)
            algorithm = received_data.get('algorithm', None)

            self._crypto_db.insert_crypto(
                package=self._package,
                hashcode=hashcode,
                algorithm=algorithm,
                init_key=None
            )
            
            if not self._suppress_messages:
                Logger.print_message(
                    level="D",
                    message=f"Cipher getInstance received\nHashcode: {hashcode}\nAlgorithm: {algorithm}\n{stack_trace}",
                    script_location=script_location
                )

        elif module == "cipher.doFinal":
            hashcode = received_data.get('hashcode', None)

            self._crypto_db.insert_crypto(
                package=self._package,
                hashcode=hashcode,
                algorithm=None,
                init_key=None
            )

            self._crypto_db.update_crypto(
                package=self._package,
                hashcode=hashcode,
                before_final=received_data.get('input', ''),
                after_final=received_data.get('output', ''),
                stack_trace=stack_trace,
                status="complete"
            )

            if not self._suppress_messages:
                data=json.dumps(dict(
                    hashcode=hashcode,
                    before_final=received_data.get('input', ''),
                    after_final=received_data.get('output', ''),
                    ), default=Logger.json_serial, indent=4, sort_keys=False)

                Logger.print_message(
                    level="D",
                    message=f"Cipher doFinal received\n{data}\n{stack_trace}",
                    script_location=script_location
                )

        elif module == "messageDigest.update":
            hashcode = received_data.get('hashcode', None)
            algorithm = received_data.get('algorithm', None)
            bInput = received_data.get('input', None)
            self._crypto_db.insert_digest(self._package, hashcode, algorithm, bInput, None, stack_trace=stack_trace)

        elif module == "messageDigest.digest":
            hashcode = received_data.get('hashcode', None)
            algorithm = received_data.get('algorithm', None)
            bInput = received_data.get('input', None)  # Se não existir teve um messageDigest.update antes
            bOutput = received_data.get('output', None)
            self._crypto_db.insert_digest(self._package, hashcode, algorithm, bInput, bOutput, stack_trace=stack_trace)

            hash_hex = ""
            if bOutput is not None:
                try:
                    if isinstance(bOutput, bytes):
                        hash_hex = ''.join('{:02x}'.format(b) for b in bOutput)
                    else:
                        hash_hex = ''.join('{:02x}'.format(b) for b in base64.b64decode(bOutput))
                except:
                    pass

            # Do not print TLS certificate verification hash
            if not self._suppress_messages:
                if 'com.android.org.conscrypt.ConscryptEngine.verifyCertificateChain' not in stack_trace:
                    Logger.print_message(
                        level="D",
                        message=f"Message digest\nAlgorithm: {algorithm}\nHash: {hash_hex}\n{stack_trace}",
                        script_location=script_location
                    )

        elif module == "KeyFactory.generatePrivate":
            #print(received_data)
            pass

        elif module == "KeyFactory.generatePublic":
            #print(received_data)
            pass

        elif module == "org.bouncycastle.asn1!init":
            #print(received_data)
            pass

        return True

    def data_event(self,
                   script_location: ScriptLocation = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        #Nothing by now
        return True





import errno
import sys
import json
import os.path
from pathlib import Path
from argparse import _ArgumentGroup, Namespace
from frida_fusion.libs.logger import Logger
from frida_fusion.libs.database import Database
from frida_fusion.module import ModuleBase
from frida_fusion.libs.scriptlocation import ScriptLocation
from frida_fusion.exceptions import SilentKillError


class Reflection(ModuleBase):

    _EXCLUSION_LIST = [
        "android.view.View.onDraw", 
        "android.graphics.Picture",
        "com.facebook.fbreact.specs.NativeAnimatedModuleSpec",
        "com.facebook.react.uimanager.BaseViewManager.setTransform",
        "com.facebook.react.uimanager.ReanimatedUIManager.updateView",
        "com.facebook.fbreact.specs.NativeStatusBarManagerAndroidSpec"
    ]

    class StalkerDB(Database):
        dbName = ""

        def __init__(self, db_name: str):
            super().__init__(
                auto_create=True,
                db_name=db_name
            )
            self.create_db()

        def create_db(self):
            super().create_db()
            conn = self.connect_to_db(check=False)

            # definindo um cursor
            cursor = conn.cursor()

            # criando a tabela (schema)
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS [reflection_stalker] (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    method TEXT NOT NULL,
                    method_b64 TEXT NULL,
                    method_name TEXT NULL,
                    target_class TEXT NULL,
                    target_tostring TEXT NULL,
                    params_count TEXT NULL,
                    params TEXT NULL,
                    error TEXT NULL,
                    return_summary TEXT NULL,
                    return_class TEXT NULL,
                    stack_trace TEXT NULL,
                    created_date datetime not null DEFAULT (datetime('now','localtime'))
                );
            """)

            conn.commit()

            # Must get the constraints
            self.get_constraints(conn)

    def __init__(self):
        super().__init__('Reflection Stalker', 'Monitor reflection calls/invoke')
        self.mod_path = str(Path(__file__).resolve().parent)
        self._stalker_db = None
        self._suppress_messages = False
        self._ignore_list = []
        self.js_file = os.path.join(self.mod_path, "reflection-stalker.js")

    def start_module(self, **kwargs) -> bool:
        if 'db_path' not in kwargs:
            raise Exception("parameter db_path not found")

        self._stalker_db = Reflection.StalkerDB(db_name=kwargs['db_path'])

    def js_files(self) -> list:
        return [
            self.js_file
        ]

    def suppress_messages(self):
        self._suppress_messages = True

    def dynamic_script(self) -> str:
        return f""
    

    def add_params(self, flags: _ArgumentGroup):
        flags.add_argument('--ignore-stalker-text',
                         dest='reflection_stalker_ignore',
                         metavar='text',
                         action='append',
                         help='Text to ignore at screen output. You can specify multiple values repeating the flag.')

    def load_from_arguments(self, args: Namespace) -> bool:
        self._ignore_list = []

        if args.reflection_stalker_ignore is None or len(args.reflection_stalker_ignore) == 0:
            return True

        self._ignore_list = list(set([
            txt.lower()
            for t1 in args.reflection_stalker_ignore
            if (txt := t1.strip()) != ''
            ]))

        return True

    def key_value_event(self,
                        script_location: ScriptLocation = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:


        if module in ["java.lang.reflect.Method!invoke!throw",
                      "java.lang.reflect.Method!invoke!call"
                      ]:
            
            # Exclusion list
            method = received_data.get('method', '')
            for em in Reflection._EXCLUSION_LIST:
                if em in method:
                    return

            params_preview = received_data.get('params_preview', '[]')
            try:
                params = json.dumps(json.loads(params_preview), default=Logger.json_serial)
            except Exception:
                params = params_preview

            return_summary = received_data.get('return_summary', None)
            try:
                return_summary = json.dumps(json.loads(return_summary), default=Logger.json_serial)
            except Exception:
                pass

            backtrace = received_data.get('backtrace_raw', None)
            try:
                if backtrace is None:
                    raise Exception()

                if isinstance(backtrace, str):
                    backtrace = json.loads(backtrace)

                bt = "Stack trace:\n    at "
                bt += '\n    at '.join([
                    ln 
                    for ln in backtrace
                    if 'dalvik.system.VMStack.getThreadStackTrace' not in ln
                    and 'java.lang.Thread.getStackTrace' not in ln
                    ])

                backtrace = bt
            except Exception:
                backtrace = received_data.get('backtrace', stack_trace)

            self._stalker_db.insert_one(
                table_name='reflection_stalker',
                stack_trace=backtrace,
                error=received_data.get('error', ''),
                method=received_data.get('method', ''),
                method_b64=received_data.get('method_b64', ''),
                method_name=received_data.get('method_name', ''),
                target_class=received_data.get('target_class', ''),
                target_tostring=received_data.get('target_tostring', ''),
                params_count=received_data.get('params_count', ''),
                params=params,
                return_summary=return_summary,
                return_class=received_data.get('return_class', '')
            )

            if not self._suppress_messages:
                txt = json.dumps(received_data, indent=4)
                for em in self._ignore_list:
                    if em in txt.lower():
                        return

                Logger.print_message(
                        level="I",
                        message=f"Reflection: {received_data.get('method', '')}",
                        script_location=script_location
                    )
                
                
                Logger.print_message(
                        level="D",
                        message=f"Reflection: {module}\n{txt}",
                        script_location=script_location
                    )

        return True

    def data_event(self,
                   script_location: ScriptLocation = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        return True





#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import os
import errno
import sys
import signal
from argparse import Namespace
from pathlib import Path

from .module import Module, ModuleManager, InternalModule, ExternalModule, LocalModule
from .libs.color import Color
from .libs.logger import Logger
from .__meta__ import __version__


class Configuration(object):
    ''' Stores configuration variables and functions for Frida Fusion. '''
    version = '0.0.0'

    initialized = False  # Flag indicating config has been initialized
    args: Namespace = {}
    debug_level = 0
    cmd_line = ''
    base_path = str(Path(__file__).resolve().parent)
    db_path = os.path.join(str(Path(".").resolve()), "fusion.db")
    enabled_modules = {}
    ignore_messages_modules = {}

    # Device vars
    device_id = None
    use_usb = False
    remote_host = None

    # App vars
    package = ''
    pid = 0

    # Scripts
    frida_scripts = None

    # General
    out_file = None
    print_timestamp = False
    use_delay = False

    @staticmethod
    def initialize():
        '''
            Sets up default initial configuration values.
            Also sets config values based on command-line arguments.
        '''

        # Only initialize this class once
        if Configuration.initialized:
            return

        Configuration.debug_level = 0  # Verbosity level.

        # Overwrite config values with arguments (if defined)
        Configuration.load_from_arguments()

        Configuration.initialized = True

    @staticmethod
    def load_from_arguments():
        ''' Sets configuration values based on Argument.args object '''
        from .args import Arguments

        config_check = 0

        sys.argv[0] = 'frida-fusion'

        Configuration.cmd_line = ' '.join([word for word in sys.argv])

        list_modules = any(['--list-modules' == word for word in sys.argv])
        #show_help = any(['-h' == word for word in sys.argv])

        if list_modules:
            mods = ModuleManager.list_modules()

            if len(mods) == 0:
                Color.pl('{!} {R}error: no modules found{R}{W}\r\n')
                sys.exit(1)

            max_name = max(iter([
                len(m.name) + 3
                for _, m in mods.items()
            ] + [15]))
            Color.pl(f"Available internal modules")
            for m in [m for _, m in mods.items() if isinstance(m, InternalModule)]:
                Color.pl(f"  {m.safe_name().ljust(max_name)} : {m.description}")

            Color.pl(f"\nAvailable external modules")
            ext_mods = [m for _, m in mods.items() if isinstance(m, ExternalModule)]
            if len(ext_mods) == 0:
                Color.pl(("  No external modules available. You can set {G}FUSION_MODULES{W} environment variable "
                          "to set an external modules Path"))
            for m in ext_mods:
                Color.pl(f"  {m.safe_name().ljust(max_name)} : {m.description}")

            print("")
            sys.exit(0)

        args = Arguments().args
        Configuration.args = args

        Color.pl('{+} {W}Startup parameters')

        if args.out_file is not None and args.out_file != '':
            Configuration.out_file = args.out_file
            try:
                with open(Configuration.out_file, 'a') as f:
                    # file opened for writing. write to it here
                    Logger.out_file = Configuration.out_file
                    f.write(Color.sc(Configuration.get_banner()) + '\n')
                    f.write(Color.sc('{+} {W}Startup parameters') + '\n')
                    pass
            except IOError as x:
                if x.errno == errno.EACCES:
                    Color.pl('{!} {R}error: could not open output file to write {O}permission denied{R}{W}\r\n')
                    sys.exit(1)
                elif x.errno == errno.EISDIR:
                    Color.pl('{!} {R}error: could not open output file to write {O}it is an directory{R}{W}\r\n')
                    sys.exit(1)
                else:
                    Color.pl('{!} {R}error: could not open output file to write{W}\r\n')
                    sys.exit(1)

        if args.app_id is None and args.pid == 0:
            Color.pl('{!} {R}error: you must specify either {O}--package{R} or {O}--attach-pid{R}{W}\r\n')
            Configuration.mandatory()

        if args.app_id is not None and args.pid > 0:
            Color.pl('{!} {R}error: you must specify just one parameter {O}--package{R} or {O}--attach-pid{R}{W}\r\n')
            Configuration.mandatory()

        Logger.pl('     {C}command line:{O} %s{W}' % Configuration.cmd_line)

        if args.app_id is not None:
            Configuration.package = args.app_id
            Logger.pl('     {C}package:{O} %s{W}' % Configuration.package)
        elif args.pid > 0:
            Configuration.pid = args.pid
            Logger.pl('     {C}process id:{O} %s{W}' % Configuration.pid)

        if args.use_usb is True:
            Configuration.use_usb = True
            Logger.pl('     {C}device:{O} USB{W}')
        elif args.device_id is not None:
            Configuration.device_id = args.device_id
            Logger.pl('     {C}device:{O} %s{W}' % Configuration.device_id)
        elif args.use_remote and args.remote_host is not None:
            Configuration.remote_host = args.remote_host
            Logger.pl('     {C}device:{O} remote %s{W}' % Configuration.remote_host)

        if Configuration.use_usb is False and Configuration.device_id is None and Configuration.remote_host is None:
            Color.pl('{!} {R}error: you must specify just one parameter {O}--usb{R}, {O}--device{R} or {O}--remote{R}{W}\r\n')
            Configuration.mandatory()

        if args.frida_scripts is not None and args.frida_scripts.strip() != "":
            if not os.path.exists(args.frida_scripts):
                Color.pl(
                    '{!} {R}error: scripts path not found{R}{W}\r\n')
                Configuration.mandatory()
            Configuration.frida_scripts = args.frida_scripts

        if Configuration.frida_scripts is None:
            Color.pl(
                '{!} {R}error: you must specify scripts path {O}--script-path{R}{W}\r\n')
            Configuration.mandatory()

        Logger.pl('     {C}scripts path:{O} %s{W}' % Configuration.frida_scripts)

        if args.delay:
            Configuration.use_delay = True

        if args.show_time:
            Configuration.print_timestamp = True
        Logger.print_timestamp = Configuration.print_timestamp

        if str(args.debug_level).upper() not in Logger.level_map.keys():
            Color.pl(
                '{!} {R}error: invalid debug level{R}{W}\r\n')
            Configuration.mandatory()

        Configuration.debug_level = Logger.level_map.get(str(args.debug_level).upper(), 0)
        Logger.debug_level = Configuration.debug_level

        Logger.pl('     {C}min debug level:{O} %s{W}' % str(args.debug_level).upper())

        mods = ModuleManager.list_modules(local_path=Path(Configuration.frida_scripts))
        if (args.enabled_modules is not None and isinstance(args.enabled_modules, list)) or \
                (args.ignore_messages_modules is not None and isinstance(args.ignore_messages_modules, list)):

            for mod in [
                m.strip()
                for md in args.enabled_modules
                for m in md.split(",")
                if m.strip() != ""
            ]:
                fm = next(iter([
                    m
                    for _, m in mods.items()
                    if m.safe_name() == mod.lower()
                ]), None)
                if fm is None:
                    Color.pl(
                        '{!} {R}error: module {O}%s{R} not found{W}\r\n' % mod)
                    sys.exit(1)

                name = fm.safe_name()
                if name not in Configuration.enabled_modules.keys():
                    Configuration.enabled_modules[name] = fm

            if args.ignore_messages_modules is not None and isinstance(args.ignore_messages_modules, list):
                for mod in [
                    m.strip()
                    for md in args.ignore_messages_modules
                    for m in md.split(",")
                    if m.strip() != ""
                ]:
                    fm = next(iter([
                        m
                        for _, m in mods.items()
                        if m.safe_name() == mod.lower()
                    ]), None)
                    if fm is None:
                        Color.pl(
                            '{!} {R}error: module {O}%s{R} not found{W}\r\n' % mod)
                        sys.exit(1)

                    name = fm.safe_name()
                    if name not in Configuration.ignore_messages_modules.keys():
                        Configuration.ignore_messages_modules[name] = fm

        # Enable user defined local modules
        for _, fm in mods.items():
            if isinstance(fm, LocalModule):
                name = fm.safe_name()
                if name not in Configuration.ignore_messages_modules.keys():
                    Configuration.ignore_messages_modules[name] = fm
                    Configuration.enabled_modules[name] = fm

        if len(Configuration.enabled_modules) > 0:
            Logger.pl('     {C}modules:{O} %s{W}' % ', '.join([
                m.name
                for _, m in Configuration.enabled_modules.items()
            ]))

        if len(Configuration.ignore_messages_modules) > 0:
            Logger.pl('     {C}ignored messages from modules:{O} %s{W}' % ', '.join([
                m.name
                for _, m in Configuration.ignore_messages_modules.items()
            ]))

        Logger.pl("")

    @staticmethod
    def get_banner():
        Configuration.version = str(__version__)

        return '''
 {P}[ {R}FRIDA {P}]{G}—o—( {O}FUSION {G})—o—{P}[ {C}MOBILE TESTS {P}] {W}{D}// v%s{W}
     {O}> {W}{D}{O}hook your mobile tests with Frida{W}
    
    ''' % Configuration.version

    @staticmethod
    def mandatory():
        Color.pl(
            '{!} {R}error: missing a mandatory option, use {O}-h{R} help{W}\r\n')
        sys.exit(1)

    @staticmethod
    def exit_gracefully(code=0):
        ''' Deletes temp and exist with the given code '''
        exit(code)

    @staticmethod
    def kill(code=0):
        ''' Deletes temp and exist with the given code '''
        os.kill(os.getpid(), signal.SIGTERM)



import json
import os.path
from pathlib import Path
from frida_fusion.libs.logger import Logger
from frida_fusion.libs.database import Database
from frida_fusion.libs.scriptlocation import ScriptLocation
from frida_fusion.module import ModuleBase


class AndroidLog(ModuleBase):
    def __init__(self):
        super().__init__('Log', 'Hook android.util.Log.* functions')
        self._package = None
        self._suppress_messages = False
        self.mod_path = str(Path(__file__).resolve().parent)

    def start_module(self, **kwargs) -> bool:
        if 'db_path' not in kwargs:
            raise Exception("parameter db_path not found")

        self._package = kwargs['package']
        return True

    def js_files(self) -> list:
        return [
            os.path.join(self.mod_path, "log.js")
        ]

    def suppress_messages(self):
        self._suppress_messages = True

    def key_value_event(self,
                        script_location: ScriptLocation = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:

        if 'android.util.Log' in module:
            level = received_data.get('level', "I")
            if level is None:
                level = "I"
            message = received_data.get('message', '')
            tag = received_data.get('tag', '')
            if tag is None or tag.strip() == '':
                tag = None
            try:
                received_data['rawargs'] = json.loads(received_data.get('rawargs', ''))
            except Exception:
                pass

            # raw_args = received_data.get('rawargs', '')
            if Logger.check_print(level="D"):
                message += f"\n{stack_trace}"

            if not self._suppress_messages:
                Logger.print_message(
                    level=level.upper(),
                    message=f"{message}",
                    script_location=ScriptLocation(
                        file_name=tag,
                        line=f"{script_location.file_name}",
                        function_name=script_location.function_name
                    ) if tag is not None else script_location
                )

        return True

    def data_event(self,
                   script_location: ScriptLocation = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        # Nothing by now
        return True



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
from __future__ import annotations

from .exceptions import SilentKillError
from .libs.color import Color
from .libs.scriptlocation import ScriptLocation

try:
    from .config import Configuration
except (ValueError, ImportError) as e:
    raise Exception('You may need to run Frida Fusion from the root directory (which includes README.md)', e)

import codecs
import time
import frida
import json
import base64
import os
import signal
import sys
import re
import threading
import queue
from pathlib import Path
from datetime import datetime

from .libs.database import Database
from .libs.logger import Logger


class Fusion(object):
    running = True
    debug = True
    print_timestamp = False
    max_filename = 28

    _bundle_pattern = re.compile(r'(fusion_bundle\.js):(\d+)')
    _script_name = Path(__file__).name
    _db_jobs = queue.Queue()

    def __init__(self):
        self.device = None
        self.session = None
        self.done = threading.Event()
        self.pid = 0
        self.script_trace = {}
        self._modules = []
        self._ignore_messages = []
        signal.signal(signal.SIGINT, self.signal_handler)

        t = threading.Thread(target=Fusion._db_worker, daemon=True)
        t.start()

    def signal_handler(self, sig, frame):
        if Logger.debug_level <= 2:
            Logger.debug_level = 2

        Fusion.running = False
        Logger.pl('\n{!} {O}interrupted, shutting down...{W}\n')
        time.sleep(0.5)
        self.done.set()

    def wait(self):
        self.done.wait()  # bloqueia até receber o set()
        try:
            self.session.detach()
        except:
            pass

        try:
            self.device.kill(self.pid)
        except:
            pass

    def get_device(self):
        try:
            if Configuration.use_usb:
                self.device = frida.get_usb_device()
            else:
                process = frida.get_device_manager()
                if Configuration.device_id is not None:
                    self.device = frida.get_device(Configuration.device_id)
                elif Configuration.remote_host is not None:
                    self.device = process.add_remote_device(Configuration.remote_host)

        except Exception as err:
            self.device = None
            Logger.print_exception(err)

        return self.device

    def translate_location(self, location: dict) -> ScriptLocation:
        if location is None or not isinstance(location, dict):
            return ScriptLocation()

        loc = ScriptLocation.parse_from_dict(location)

        if loc.file_name != "fusion_bundle.js":
            return loc

        return next(iter([
            ScriptLocation(
                file_name=k,
                function_name=loc.function_name,
                line=str(1 + loc.get_int_line() - v[0])
            )
            for k, v in self.script_trace.items()
            if v[0] <= loc.get_int_line() <= v[1]
        ]), loc)

    def load_all_scripts(self):
        self.script_trace = {}
        offset = 1
        line_cnt = 0

        Logger.pl("{*} Loading FridaFusion helpers script")
        src = ""
        try:
            src += self.sanitize_js(
                open(os.path.join(
                    Configuration.base_path, "libs", 'helpers.js'),
                    'r', encoding='utf-8').read())
            src += "\n\n"
        except Exception as e:
            Logger.pl('\n{!} {R}Error loading helpers.js:{O} %s{W}' % str(e))
            sys.exit(1)

        line_cnt = len(src.split("\n")) - 1

        self.script_trace['helpers.js'] = (offset, line_cnt)
        offset += line_cnt

        files_js = []

        for m in self._modules:
            files_js += m.js_files()
            dyn = m.dynamic_script()
            if dyn is not None and dyn.strip(" \r\n") != "":
                dyn += "\n\n"
                line_cnt = len(dyn.split("\n")) - 1
                self.script_trace[f"dyn_{m.safe_name()}.js"] = (offset, offset + line_cnt)
                offset += line_cnt
                src += dyn

        if os.path.isfile(Configuration.frida_scripts):
            if Path(Configuration.frida_scripts).suffix.lower() == ".js":
                files_js += [Configuration.frida_scripts]
        else:
            files_js += [
                os.path.join(Configuration.frida_scripts, f)
                for f in sorted(os.listdir(Configuration.frida_scripts))
                if f.endswith(".js")
            ]

        # Keep unique files
        # Do not use list(set(files_js)) because it will lose the order of modules
        done: set[str] = set()
        for file_path in files_js:
            if file_path in done:
                continue

            done.add(file_path)

            file_name = Path(file_path).name
            file_data = self.sanitize_js(open(file_path, 'r', encoding='utf-8').read())
            if '#NOLOAD' in file_data:
                Logger.pl('{!} {O}Alert:{W} {G}#NOLOAD{W} tag found at {G}%s{W}, ignoring file.{W}' % str(file_name))
            else:
                Logger.pl("{*} Loading script file " + file_name)
                for r in ["*", "-", "+", "!"]:
                    file_data = file_data.replace(f"console.log('[{r}] ", f"fusion_sendMessage('{r}', '")
                    file_data = file_data.replace(f'console.log("[{r}] ', f'fusion_sendMessage("{r}", "')
                    file_data = file_data.replace(f"console.log('[{r}]", f"fusion_sendMessage('{r}', '")
                    file_data = file_data.replace(f'console.log("[{r}]', f'fusion_sendMessage("{r}", "')

                file_data = re.sub(r'(?<!\w)send\(', 'fusion_Send(', file_data)

                file_data = file_data.replace(f'console.log(', f'fusion_sendMessage("I", ')
                file_data += "\n\n"

                line_cnt = len(file_data.split("\n")) - 1

                self.script_trace[file_name] = (offset, offset + line_cnt - 1)
                offset += line_cnt

                src += file_data

                if len(file_name) > Fusion.max_filename:
                    Fusion.max_filename = len(file_name)

            Logger.filename_col_len = Fusion.max_filename

        try:
            s = self.session.create_script(src, name="fusion_bundle")
            s.on("message", self.make_handler("fusion_bundle.js"))  # register the message handler
            s.load()
        except Exception as err:
            try:
                from traceback import format_exc
                err_txt = 'Error:{O} %s{W}' % str(err)
                err_txt += '\n{O}Full stack trace below\n'
                err_txt += format_exc().strip()

                err_txt = err_txt.replace('\n', '\n{W}   ')
                err_txt = err_txt.replace('  File', '{W}{D}File')
                err_txt = err_txt.replace('  Exception: ', '{R}Exception: {O}')

                pattern = re.compile(r'script\(line (\d+)\):')
                matches = [
                    (
                        m.group(0),
                        self.translate_location(dict(
                            file_name="fusion_bundle.js",
                            line=m.group(1),
                        ))
                    )
                    for m in pattern.finditer(err_txt)
                ]
                for m in matches:
                    err_txt = err_txt.replace(m[0], f"{m[1].file_name}(line {m[1].line})")

                Logger.pl(err_txt)
                print("")
                sys.exit(1)
            except Exception as e2:
                Logger.print_exception(e2)
                print("")
                sys.exit(1)

    def attach(self, pid: int):
        Fusion.running = True
        self.pid = pid

        if self.session is not None:
            try:
                self.session.off("detached", self.on_detached)
            except:
                pass
            self.session = None

        self.session = self.device.attach(self.pid)
        self.session.on("detached", self.on_detached)

        Logger.pl("{+} Starting frida scripts")
        self.load_all_scripts()
        self.device.resume(self.pid)

    def std_spawn(self):
        Fusion.running = True

        if self.session is not None:
            try:
                self.session.off("detached", self.on_detached)
            except:
                pass
            self.session = None

        self.pid = self.device.spawn([Configuration.package])
        self.session = self.device.attach(self.pid)
        self.session.on("detached", self.on_detached)

        Logger.pl("{+} Starting frida scripts")
        self.load_all_scripts()
        self.device.resume(self.pid)

    def wait_spawn(self):
        Fusion.running = True

        if self.session is not None:
            try:
                self.session.off("detached", self.on_detached)
            except:
                pass
            self.session = None

        self.pid = self.device.spawn([Configuration.package])
        self.device.resume(self.pid)

        time.sleep(0.2)  # Without it Java.perform silently fails

        self.session = self.device.attach(self.pid)
        self.session.on("detached", self.on_detached)

        Logger.pl("{+} Starting frida scripts")
        self.load_all_scripts()

    def make_handler(self, script_name):
        def handler(message, payload):

            if not Fusion.running:
                return

            if message["type"] == "send":
                try:
                    script_location = ScriptLocation()
                    jData = message.get("payload", {})
                    if isinstance(jData, str):
                        jData = json.loads(message["payload"])

                    # Check another payload level
                    p1 = jData.get("payload", None)
                    if p1 is not None:
                        location = jData.get("location", None)
                        jData = jData.get("payload", {})
                        script_location = self.translate_location(location)

                    if isinstance(jData, str):
                        msg = jData
                        try:
                            msg = base64.b64encode(jData.encode("UTF-8"))
                        except Exception:
                            pass

                        jData = {
                            "type": "message",
                            "level": "I",
                            "message": msg
                        }

                    if script_location.file_name == "<unknown>":
                        script_location.file_name = script_name

                    mType = jData.get('type', '').lower()
                    mLevel = jData.get('level', None)
                    if mType == "message":

                        if script_location.file_name in self._ignore_messages:
                            return

                        msg = jData.get('message', '')
                        try:
                            msg = base64.b64decode(msg).decode("UTF-8")
                        except:
                            pass

                        self.print_message_inst(mLevel, msg, script_location=script_location)

                    elif mType == "key_value_data":
                        self.print_message_inst("V", "RAW JSON:\n    %s" % (
                            json.dumps(jData, indent=4).replace("\n", "\n    ")
                        ), script_location=script_location)

                        stack_trace = jData.get('stack_trace', '')
                        try:
                            stack_trace = base64.b64decode(stack_trace).decode("UTF-8")
                        except:
                            pass
                        self.insert_history('frida', json.dumps(jData), stack_trace)

                        if len(Configuration.enabled_modules) > 0:

                            received_data = {
                                k.lower(): v
                                for item in jData.get('data', [])
                                if isinstance(item, dict)
                                if (k := item.get("key", None)) is not None
                                   and (v := item.get("value", None)) is not None
                            }

                            self._raise_key_value_event(
                                script_location=script_location,
                                stack_trace=stack_trace,
                                module=jData.get('module', None),
                                received_data=received_data
                            )

                    # Legacy
                    elif mType == "data":
                        self.print_message_inst("V", "RAW JSON:\n    %s" % (
                            json.dumps(jData, indent=4).replace("\n", "\n    ")
                        ), script_location=script_location)

                        stack_trace = jData.get('stack_trace', '')
                        try:
                            stack_trace = base64.b64decode(stack_trace).decode("UTF-8")
                        except:
                            pass
                        self.insert_history('frida', json.dumps(jData), stack_trace)

                        self._raise_data_event(
                            script_location=script_location,
                            stack_trace=stack_trace,
                            received_data=json.dumps(jData)
                        )

                    elif mType == "native-exception":
                        self.insert_history('frida', json.dumps(jData))
                        if self.check_frida_native_exception(jData):
                            Logger.pl(self.format_frida_native_exception(jData))

                    elif mType == "java-uncaught":
                        self.insert_history('frida', json.dumps(jData))
                        self.print_message_inst("E", jData.get('stack', ''), script_location=script_location)

                    else:
                        self.print_message_inst(mLevel, message, script_location=script_location)

                except SilentKillError as sk:
                    skm = str(sk)

                    self.print_message_inst("D", "Silent kill requested",
                                            script_location=Logger.get_caller_info(stack_index=1))
                    Fusion.running = False
                    time.sleep(0.2)
                    if skm != "":
                        Logger.pl(f'\n{skm}')
                    Logger.pl('\n{+} {O}Exiting...{O}{W}')
                    self.done.set()

                except Exception as err:
                    script_location = ScriptLocation(file_name=Fusion._script_name)
                    self.print_message_inst("E", message, script_location=script_location)
                    self.print_message_inst("E", payload, script_location=script_location)
                    self.print_exception(err)

            else:
                script_location = ScriptLocation.parse_from_dict(message)
                try:
                    if message["type"] == "error":
                        description = message.get('description', '')
                        if description is not None and description.strip() != "":
                            description += "\n"

                        stack = "Stack trace:\n"
                        stack += message.get('stack', '')

                        matches = [
                            (
                                m.group(0),
                                self.translate_location(dict(
                                    file_name=m.group(1),
                                    line=m.group(2),
                                ))
                            )
                            for m in Fusion._bundle_pattern.finditer(stack)
                        ]
                        for m in matches:
                            stack = stack.replace(m[0], f"{m[1].file_name}:{m[1].line}")

                        if script_location.file_name == "fusion_bundle.js" and len(matches) >= 1:
                            script_location.file_name = matches[0][1].file_name
                            script_location.line = matches[0][1].line

                        elif script_location.file_name == "class-factory.js" and len(matches) == 1:
                            script_location.file_name = matches[0][1].file_name
                            script_location.line = matches[0][1].line

                        self.insert_history('frida', json.dumps({
                            "type": "error",
                            "description": description,
                            "stack": stack
                        }))

                        self.print_message_inst("F", description + stack,
                                                script_location=script_location)
                        Fusion.running = False
                        time.sleep(0.2)
                        Logger.pl('\n{+} {O}Exiting...{O}{W}')
                        self.done.set()
                    else:
                        self.print_message_inst("I", message, script_location=script_location)
                        self.print_message_inst("I", payload, script_location=script_location)
                except Exception as e:
                    self.print_message_inst("I", message, script_location=script_location)
                    self.print_message_inst("I", payload, script_location=script_location)
                    self.print_exception(e)

        return handler

    def on_detached(self, reason, crash):
        Logger.pl('\n{!} {R}DETACHED:{O} reason=%s{W}' % str(reason))
        if crash is not None and isinstance(crash, dict):
            # crash é um dict com info de sinal, endereço, etc. quando disponível
            try:
                Logger.pl("[CRASH] details: " + json.dumps(crash))
            except:
                Logger.pl("[CRASH] details: " + str(crash))
        elif crash is not None:
            Logger.pl("[CRASH] details: " + str(crash))

        Logger.pl("")
        self.done.set()

    def _replace_location(self, message: str) -> str:
        try:
            matches = [
                (
                    m.group(0),
                    self.translate_location(dict(
                        file_name=m.group(1),
                        line=m.group(2),
                    ))
                )
                for m in Fusion._bundle_pattern.finditer(message)
            ]
            for m in matches:
                message = message.replace(m[0], f"{m[1].file_name}:{m[1].line}")
        except Exception as e:
            print(e)
            pass

        return message

    def _raise_key_value_event(self,
                               script_location: ScriptLocation = None,
                               stack_trace: str = None,
                               module: str = None,
                               received_data: dict = None):
        for m in self._modules:
            try:
                m.key_value_event(
                    script_location=script_location,
                    stack_trace=stack_trace,
                    module=module,
                    received_data=received_data
                )
            except SilentKillError as ske:
                raise ske
            except Exception as err:
                self.print_exception(
                    err,
                    script_location=Logger.get_error_info_from_format_exc(stack_index=-1)
                )

    def _raise_data_event(self,
                          script_location: ScriptLocation = None,
                          stack_trace: str = None,
                          received_data: str = None):
        for m in self._modules:
            try:
                m.data_event(
                    script_location=script_location,
                    stack_trace=stack_trace,
                    received_data=received_data
                )
            except SilentKillError as ske:
                raise ske
            except Exception as err:
                self.print_exception(
                    err,
                    script_location=Logger.get_error_info_from_format_exc(stack_index=-1)
                )

    def print_message_inst(self, level: str = "*", message: str = "",
                           script_location: ScriptLocation = None):

        return type(self)._print_message(
            level=level,
            message=self._replace_location(message),
            script_location=script_location
            )

    @classmethod
    def print_message(cls, level: str = "*", message: str = "",
                      script_location: ScriptLocation = None):
        return cls._print_message(
            level=level,
            message=message,
            script_location=script_location
            )

    @classmethod
    def _print_message(cls, level: str = "*", message: str = "",
                       script_location: ScriptLocation = None):

        if Fusion.running is False and Logger.debug_level >= 2:
            return

        if script_location is None:
            script_location = ScriptLocation(
                file_name=Fusion._script_name
            )

        Logger.print_message(
            level=level,
            message=message,
            script_location=script_location,
            filename_col_len=Fusion.max_filename
        )

    @classmethod
    def insert_history(cls,  source: str, data: str, stack_trace: str = ''):
        Fusion._db_jobs.put(dict(
            source=source,
            data=data,
            stack_trace=stack_trace
        ))

    @classmethod
    def _db_worker(cls):
        db = Database(auto_create=True, db_name=Configuration.db_path)

        while Fusion.running:
            kwargs = Fusion._db_jobs.get()
            db.insert_history(**kwargs)

    @classmethod
    def print_exception(cls, err, script_location: ScriptLocation = None):
        from traceback import format_exc
        err_txt = 'Error:{O} %s{W}' % str(err)
        err_txt += '\n{O}Full stack trace below\n'
        err_txt += format_exc().strip()

        err_txt = err_txt.replace('\n', '\n{W}   ')
        err_txt = err_txt.replace('  File', '{W}{D}File')
        err_txt = err_txt.replace('  Exception: ', '{R}Exception: {O}')

        Logger.print_message(
            level="E",
            message=Color.s(err_txt),
            filename_col_len=Fusion.max_filename,
            script_location=script_location if script_location is not None else Logger.get_caller_info(stack_index=2)
        )

    @classmethod
    def check_frida_native_exception(cls, evt: dict) -> bool:
        d = evt.get('details', {})
        mem = d.get('memory', {}) or {}
        bt = d.get('backtrace', []) or []
        # if mem.get('address','') == "0x0" and len(bt) > 0 and ('boot-core' in bt[0] or 'boot-framework' in bt[0]):
        if len(bt) > 0 and ('boot-core' in bt[0] or 'boot-framework' in bt[0]):
            return False

        return True

    @classmethod
    def format_frida_native_exception(cls, evt: dict) -> str:
        # ANSI
        RED = "\033[31m"
        GREEN = "\033[32m"
        YELLOW = "\033[33m"
        BLUE = "\033[34m"
        MAGENTA = "\033[35m"
        CYAN = "\033[36m"
        BOLD = "\033[1m"
        DIM = "\033[2m"
        RESET = "\033[0m"

        d = evt.get('details', {})
        msg = d.get('message', '')
        ety = d.get('type', '')
        adr = d.get('address', '')
        mem = d.get('memory', {}) or {}
        ctx = d.get('context', {}) or {}
        ncx = d.get('nativeContext', '')
        bt = d.get('backtrace', []) or []  # lista já formatada vinda do JS

        def reg_line(keys):
            return "  " + "   ".join(
                f"{CYAN}{k.rjust(3)}{RESET}: {MAGENTA}{ctx.get(k, '').ljust(18)}{RESET}" for k in keys)

        regs_order = [
            'pc', 'lr', 'sp', 'fp',
            'x0', 'x1', 'x2', 'x3', 'x4', 'x5', 'x6', 'x7', 'x8', 'x9', 'x10', 'x11', 'x12', 'x13', 'x14', 'x15',
            'x16', 'x17', 'x18', 'x19', 'x20', 'x21', 'x22', 'x23', 'x24', 'x25', 'x26', 'x27', 'x28'
        ]

        lines = [""]
        lines.append(f"{BOLD}{RED}===== Native Exception ====={RESET}")
        lines.append(f"{YELLOW}Type:{RESET} {ety}")
        lines.append(f"{YELLOW}Message:{RESET} {msg}")
        lines.append(f"{YELLOW}Address:{RESET} {GREEN}{adr}{RESET}\n")

        if mem:
            lines.append(f"{BOLD}Memory:{RESET}")
            lines.append(f"  {CYAN}operation{RESET}: {mem.get('operation', '')}")
            lines.append(f"  {CYAN}address{RESET}:   {MAGENTA}{mem.get('address', '')}{RESET}\n")

        lines.append(f"{BOLD}Context (ARM64):{RESET}")

        group = []
        for r in regs_order[4:]:
            group.append(r)
            if len(group) == 4:
                lines.append(reg_line(group))
                group = []
        if group:
            lines.append(reg_line(group))

        lines.append(reg_line(['lr', 'sp', 'pc', 'fp']))

        lines.append(f"\n{YELLOW}nativeContext:{RESET} {MAGENTA}{ncx}{RESET}")

        # Backtrace (se disponível)
        if bt:
            lines.append(f"\n{BOLD}{BLUE}Backtrace:{RESET}")
            for frame in bt:
                # Heurística simples de cor: endereço/offset em magenta, módulo em verde
                # frame já vem no formato " 0  func (module+0xOFF)"
                # vamos apenas aplicar cores mantendo o texto
                try:
                    func_part, rest = frame.split(" (", 1)
                    mod_part = rest.rstrip(")")
                    # tenta separar "module+0xOFF"
                    if "+" in mod_part:
                        mod_name, off = mod_part.split("+", 1)
                        colored = f"{DIM}{func_part}{RESET} ({GREEN}{mod_name}{RESET}+{MAGENTA}{off}{RESET})"
                    else:
                        colored = f"{DIM}{func_part}{RESET} ({GREEN}{mod_part}{RESET})"
                except Exception:
                    colored = frame  # fallback sem cores se parsing falhar
                lines.append("  " + colored)

        lines.append(f"{BOLD}{RED}============================{RESET}")
        return "\n".join(lines)

    @classmethod
    def sanitize_js(cls, s: str) -> str:
        s = s.lstrip('\ufeff')          # remove BOM
        s = s.replace('\u2028', '\n')   # Unicode line separator
        s = s.replace('\u2029', '\n')   # Unicode paragraph separator
        s = s.replace('\u00A0', ' ')    # non-breaking space
        s = s.replace('\u200B', '')     # zero-width space
        return s

    def main(self):
        Logger.pl(Configuration.get_banner())
        Configuration.initialize()

        try:
            print(f" 🛠️  Starting Frida Fusion instrumentation")
            timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            Logger.pl('{+} {C}Start time {O}%s{W}' % timestamp)

            self._modules = [
                m.create_instance()
                for _, m in Configuration.enabled_modules.items()
            ]

            if len(self._modules) > 0:
                Logger.pl("{+} Starting selected modules")
                for m in self._modules:
                    m.load_from_arguments(Configuration.args)
                    m.start_module(
                        package=Configuration.package,
                        db_path=Configuration.db_path
                    )
                    if m.safe_name() in Configuration.ignore_messages_modules.keys():
                        m.suppress_messages()

                self._ignore_messages = [
                    Path(f).name
                    for _, md in Configuration.ignore_messages_modules.items()
                    if (m := next(iter([
                        mi
                        for mi in self._modules
                        if mi.name == md.name
                    ]), None)) is not None
                    for f in m.js_files()
                ]

            self.get_device()
            if self.device is not None:
                if Configuration.pid > 0:
                    self.attach(Configuration.pid)
                else:
                    if Configuration.use_delay:
                        self.wait_spawn()
                    else:
                        self.std_spawn()

                self.wait()
            else:
                print(' ')
                sys.exit(4)

            timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            Logger.pl('{+} {C}End time {O}%s{W}' % timestamp)
            print(' ')
            sys.exit(0)

        except Exception as e:
            Logger.pl('\n{!} {R}Error:{O} %s{W}' % str(e))

            if Configuration.debug_level <= 1:
                Logger.pl('\n{!} {O}Full stack trace below')
                from traceback import format_exc
                err = format_exc().strip()
                err = err.replace('\n', '\n{W}{!} {W}   ')
                err = err.replace('  File', '{W}{D}File')
                err = err.replace('  Exception: ', '{R}Exception: {O}')
                Logger.pl('\n{!}    ' + err)

            Logger.pl('\n{!} {R}Exiting{W}\n')
            sys.exit(2)

        except KeyboardInterrupt as e:
            sys.exit(3)


def run():
    # Explicitly changing the stdout encoding format
    if sys.stdout.encoding is None:
        sys.stdout = codecs.getwriter('latin-1')(sys.stdout)

    o = Fusion()
    o.main()

